from deepchem.models.torch_models.hf_models import HuggingFaceModel
from transformers.models.roberta.modeling_roberta import (
    RobertaConfig, RobertaForMaskedLM, RobertaForSequenceClassification)
from transformers.models.roberta.tokenization_roberta_fast import \
    RobertaTokenizerFast
from transformers.modeling_utils import PreTrainedModel


class Chemberta(HuggingFaceModel):
    """Chemberta Model

    Chemberta is a transformer style model for learning on SMILES strings.
    The model architecture is based on the RoBERTa architecture. The model
    has can be used for both pretraining an embedding and finetuning for
    downstream applications.

    The model supports two types of pretraining tasks - pretraining via masked language
    modeling and pretraining via multi-task regression. To pretrain via masked language
    modeling task, use task = `mlm` and for pretraining via multitask regression task,
    use task = `mtr`. The model supports the regression, classification and multitask
    regression finetuning tasks and they can be specified using `regression`, `classification`
    and `mtr` as arguments to the `task` keyword during model initialisation.

    The model uses a tokenizer To create input tokens for the models from the SMILES strings.
    The default tokenizer model is a byte-pair encoding tokenizer trained on PubChem10M dataset
    and loaded from huggingFace model hub (https://huggingface.co/seyonec/PubChem10M_SMILES_BPE_60k).


    Parameters
    ----------
    task: str
        The task defines the type of learning task in the model. The supported tasks are
         - `mlm` - masked language modeling commonly used in pretraining
         - `mtr` - multitask regression - a task used for both pretraining base models and finetuning
         - `regression` - use it for regression tasks, like property prediction
         - `classification` - use it for classification tasks
    tokenizer_path: str
        Path containing pretrained tokenizer used to tokenize SMILES string for model inputs. The tokenizer path can either be a huggingFace tokenizer model or a path in the local machine containing the tokenizer.
    n_tasks: int, default 1
        Number of prediction targets for a multitask learning model

    Example
    -------
    >>> import os
    >>> import tempfile
    >>> tempdir = tempfile.mkdtemp()

    >>> # preparing dataset
    >>> import pandas as pd
    >>> import deepchem as dc
    >>> smiles = ["CCN(CCSC)C(=O)N[C@@](C)(CC)C(F)(F)F","CC1(C)CN(C(=O)Nc2cc3ccccc3nn2)C[C@@]2(CCOC2)O1"]
    >>> labels = [3.112,2.432]
    >>> df = pd.DataFrame(list(zip(smiles, labels)), columns=["smiles", "task1"])
    >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
    ...     df.to_csv(tmpfile.name)
    ...     loader = dc.data.CSVLoader(["task1"], feature_field="smiles", featurizer=dc.feat.DummyFeaturizer())
    ...     dataset = loader.create_dataset(tmpfile.name)

    >>> # pretraining
    >>> from deepchem.models.torch_models.chemberta import Chemberta
    >>> pretrain_model_dir = os.path.join(tempdir, 'pretrain-model')
    >>> tokenizer_path = "seyonec/PubChem10M_SMILES_BPE_60k"
    >>> pretrain_model = Chemberta(task='mlm', model_dir=pretrain_model_dir, tokenizer_path=tokenizer_path)  # mlm pretraining
    >>> pretraining_loss = pretrain_model.fit(dataset, nb_epoch=1)

    >>> # finetuning in regression mode
    >>> finetune_model_dir = os.path.join(tempdir, 'finetune-model')
    >>> finetune_model = Chemberta(task='regression', model_dir=finetune_model_dir, tokenizer_path=tokenizer_path)
    >>> finetune_model.load_from_pretrained(pretrain_model_dir)
    >>> finetuning_loss = finetune_model.fit(dataset, nb_epoch=1)

    >>> # prediction and evaluation
    >>> result = finetune_model.predict(dataset)
    >>> eval_results = finetune_model.evaluate(dataset, metrics=dc.metrics.Metric(dc.metrics.mae_score))


    Reference
    ---------
    .. Chithrananda, S., Grand, G., & Ramsundar, B. (2020). Chemberta: Large-scale self-supervised pretraining for molecular property prediction. arXiv preprint arXiv:2010.09885.
    .. Ahmad, Walid, et al. "Chemberta-2: Towards chemical foundation models." arXiv preprint arXiv:2209.01712 (2022).
    """

    def __init__(self,
                 task: str,
                 tokenizer_path: str = 'seyonec/PubChem10M_SMILES_BPE_60k',
                 n_tasks: int = 1,
                 **kwargs):
        self.n_tasks = n_tasks
        tokenizer = RobertaTokenizerFast.from_pretrained(tokenizer_path)
        model: PreTrainedModel
        chemberta_config = RobertaConfig(vocab_size=tokenizer.vocab_size)
        if task == 'mlm':
            model = RobertaForMaskedLM(chemberta_config)
        elif task == 'mtr':
            chemberta_config.problem_type = 'regression'
            chemberta_config.num_labels = n_tasks
            model = RobertaForSequenceClassification(chemberta_config)
        elif task == 'regression':
            chemberta_config.problem_type = 'regression'
            chemberta_config.num_labels = n_tasks
            model = RobertaForSequenceClassification(chemberta_config)
        elif task == 'classification':
            model = RobertaForSequenceClassification(chemberta_config)
        else:
            raise ValueError('invalid task specification')

        super(Chemberta, self).__init__(model=model,
                                        task=task,
                                        tokenizer=tokenizer,
                                        **kwargs)



import numpy as np
import tensorflow as tf
from deepchem.data import NumpyDataset
from deepchem.feat import CircularFingerprint
from deepchem.models import KerasModel
from deepchem.models.losses import HingeLoss
from tensorflow.keras.layers import Input, Dense, Dropout, Lambda


class ScScoreModel(KerasModel):
    """
    https://pubs.acs.org/doi/abs/10.1021/acs.jcim.7b00622
    Several definitions of molecular complexity exist to facilitate prioritization
    of lead compounds, to identify diversity-inducing and complexifying reactions,
    and to guide retrosynthetic searches. In this work, we focus on synthetic
    complexity and reformalize its definition to correlate with the expected number
    of reaction steps required to produce a target molecule, with implicit knowledge
    about what compounds are reasonable starting materials. We train a neural
    network model on 12 million reactions from the Reaxys database to impose a
    pairwise inequality constraint enforcing the premise of this definition: that on
    average, the products of published chemical reactions should be more
    synthetically complex than their corresponding reactants. The learned metric
    (SCScore) exhibits highly desirable nonlinear behavior, particularly in
    recognizing increases in synthetic complexity throughout a number of linear
    synthetic routes.

    Our model here actually uses hingeloss instead of the shifted relu loss in
    https://github.com/connorcoley/scscore.

    This could cause issues differentiation issues with compounds that are "close"
    to each other in "complexity"

    """

    def __init__(self,
                 n_features,
                 layer_sizes=[300, 300, 300],
                 dropouts=0.0,
                 **kwargs):
        """
        Parameters
        ----------
        n_features: int
            number of features per molecule
        layer_sizes: list of int
            size of each hidden layer
        dropouts: int
            droupout to apply to each hidden layer
        kwargs
            This takes all kwards as TensorGraph
        """
        self.n_features = n_features
        self.layer_sizes = layer_sizes
        self.dropout = dropouts

        m1_features = Input(shape=(self.n_features,))
        m2_features = Input(shape=(self.n_features,))
        prev_layer1 = m1_features
        prev_layer2 = m2_features
        for layer_size in self.layer_sizes:
            layer = Dense(layer_size, activation=tf.nn.relu)
            prev_layer1 = layer(prev_layer1)
            prev_layer2 = layer(prev_layer2)
            if self.dropout > 0.0:
                prev_layer1 = Dropout(rate=self.dropout)(prev_layer1)
                prev_layer2 = Dropout(rate=self.dropout)(prev_layer2)

        readout_layer = Dense(1)
        readout_m1 = readout_layer(prev_layer1)
        readout_m2 = readout_layer(prev_layer2)
        outputs = [
            Lambda(lambda x: tf.sigmoid(x) * 4 + 1)(readout_m1),
            Lambda(lambda x: tf.sigmoid(x) * 4 + 1)(readout_m2),
            Lambda(lambda x: x[0] - x[1])([readout_m1, readout_m2])
        ]
        output_types = ['prediction', 'prediction', 'loss']
        model = tf.keras.Model(inputs=[m1_features, m2_features],
                               outputs=outputs)
        super(ScScoreModel, self).__init__(model,
                                           HingeLoss(),
                                           output_types=output_types,
                                           **kwargs)

    def default_generator(self,
                          dataset,
                          epochs=1,
                          mode='fit',
                          deterministic=True,
                          pad_batches=True):
        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 ids_b) in dataset.iterbatches(batch_size=self.batch_size,
                                               deterministic=deterministic,
                                               pad_batches=pad_batches):
                yield ([X_b[:, 0], X_b[:, 1]], [y_b], [w_b])

    def predict_mols(self, mols):
        featurizer = CircularFingerprint(size=self.n_features,
                                         radius=2,
                                         chiral=True)
        features = np.expand_dims(featurizer.featurize(mols), axis=1)
        features = np.concatenate([features, features], axis=1)
        ds = NumpyDataset(features, None, None, None)
        return self.predict(ds)[0][:, 0]



import unittest

from deepchem.utils.molecule_feature_utils import one_hot_encode
from deepchem.utils.molecule_feature_utils import get_atom_type_one_hot
from deepchem.utils.molecule_feature_utils import construct_hydrogen_bonding_info
from deepchem.utils.molecule_feature_utils import get_atom_hydrogen_bonding_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_hybridization_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_total_num_Hs_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_is_in_aromatic_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_chirality_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_formal_charge
from deepchem.utils.molecule_feature_utils import get_atom_formal_charge_one_hot

from deepchem.utils.molecule_feature_utils import get_atom_partial_charge
from deepchem.utils.molecule_feature_utils import get_atom_total_degree_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_implicit_valence_one_hot
from deepchem.utils.molecule_feature_utils import get_atom_explicit_valence_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_type_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_is_in_same_ring_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_is_conjugated_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_stereo_one_hot
from deepchem.utils.molecule_feature_utils import get_bond_graph_distance_one_hot


class TestGraphConvUtils(unittest.TestCase):

    def setUp(self):
        from rdkit import Chem
        self.mol = Chem.MolFromSmiles("CN=C=O")  # methyl isocyanate
        self.mol_copper_sulfate = Chem.MolFromSmiles("[Cu+2].[O-]S(=O)(=O)[O-]")
        self.mol_benzene = Chem.MolFromSmiles("c1ccccc1")
        self.mol_s_alanine = Chem.MolFromSmiles("N[C@@H](C)C(=O)O")

    def test_one_hot_encode(self):
        # string set
        assert one_hot_encode("a", ["a", "b", "c"]) == [1.0, 0.0, 0.0]
        # integer set
        assert one_hot_encode(2, [0.0, 1, 2]) == [0.0, 0.0, 1.0]
        # include_unknown_set is False
        assert one_hot_encode(3, [0.0, 1, 2]) == [0.0, 0.0, 0.0]
        # include_unknown_set is True
        assert one_hot_encode(3, [0.0, 1, 2], True) == [0.0, 0.0, 0.0, 1.0]

    def test_get_atom_type_one_hot(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_type_one_hot(atoms[0])
        assert one_hot == [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        # check unknown atoms
        atoms = self.mol_copper_sulfate.GetAtoms()
        assert atoms[0].GetSymbol() == "Cu"
        one_hot = get_atom_type_one_hot(atoms[0])
        assert one_hot == [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.0]
        one_hot = get_atom_type_one_hot(atoms[0], include_unknown_set=False)
        assert one_hot == [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        # check original set
        atoms = self.mol.GetAtoms()
        assert atoms[1].GetSymbol() == "N"
        original_set = ["C", "O", "N"]
        one_hot = get_atom_type_one_hot(atoms[1], allowable_set=original_set)
        assert one_hot == [0.0, 0.0, 1.0, 0.0]

    def test_construct_hydrogen_bonding_info(self):
        info = construct_hydrogen_bonding_info(self.mol)
        assert isinstance(info, list)
        assert isinstance(info[0], tuple)
        # Generally, =O behaves as an electron acceptor
        assert info[0] == (3, "Acceptor")

    def test_get_atom_hydrogen_bonding_one_hot(self):
        info = construct_hydrogen_bonding_info(self.mol)
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_hydrogen_bonding_one_hot(atoms[0], info)
        assert one_hot == [0.0, 0.0]

        assert atoms[3].GetSymbol() == "O"
        one_hot = get_atom_hydrogen_bonding_one_hot(atoms[3], info)
        assert one_hot == [0.0, 1.0]

    def test_get_atom_is_in_aromatic_one_hot(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_is_in_aromatic_one_hot(atoms[0])
        assert one_hot == [0.0]

        atoms = self.mol_benzene.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_is_in_aromatic_one_hot(atoms[0])
        assert one_hot == [1.0]

    def test_get_atom_hybridization_one_hot(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_hybridization_one_hot(atoms[0])
        assert one_hot == [0.0, 0.0, 1.0]

    def test_get_atom_total_num_Hs_one_hot(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_total_num_Hs_one_hot(atoms[0])
        assert one_hot == [0.0, 0.0, 0.0, 1.0, 0.0, 0.0]
        assert atoms[3].GetSymbol() == "O"
        one_hot = get_atom_total_num_Hs_one_hot(atoms[3])
        assert one_hot == [1.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    def test_get_atom_chirality_one_hot(self):
        atoms = self.mol_s_alanine.GetAtoms()
        assert atoms[0].GetSymbol() == "N"
        one_hot = get_atom_chirality_one_hot(atoms[0])
        assert one_hot == [0.0, 0.0]
        assert atoms[1].GetSymbol() == "C"
        one_hot = get_atom_chirality_one_hot(atoms[1])
        assert one_hot == [0.0, 1.0]

    def test_get_atom_formal_charge(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        formal_charge = get_atom_formal_charge(atoms[0])
        assert formal_charge == [0.0]

    def test_get_atom_formal_charge_one_hot(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        formal_charge = get_atom_formal_charge_one_hot(atoms[0])
        assert formal_charge == [0.0, 0.0, 1.0, 0.0, 0.0, 0.0]

    def test_get_atom_partial_charge(self):
        from rdkit.Chem import AllChem
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        with self.assertRaises(KeyError):
            get_atom_partial_charge(atoms[0])

        # we must compute partial charges before using `get_atom_partial_charge`
        AllChem.ComputeGasteigerCharges(self.mol)
        partial_charge = get_atom_partial_charge(atoms[0])
        assert len(partial_charge) == 1.0
        assert isinstance(partial_charge[0], float)

    def test_get_atom_total_degree_one_hot(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_total_degree_one_hot(atoms[0])
        assert one_hot == [0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0]

        assert atoms[3].GetSymbol() == "O"
        one_hot = get_atom_total_degree_one_hot(atoms[3])
        assert one_hot == [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    def test_get_atom_implicit_valence_one_hot(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_implicit_valence_one_hot(atoms[0])
        assert one_hot == [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0]

        assert atoms[3].GetSymbol() == "O"
        one_hot = get_atom_implicit_valence_one_hot(atoms[3])
        assert one_hot == [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    def test_get_atom_explicit_valence_one_hot(self):
        atoms = self.mol.GetAtoms()
        assert atoms[0].GetSymbol() == "C"
        one_hot = get_atom_explicit_valence_one_hot(atoms[0])
        assert one_hot == [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]

        assert atoms[3].GetSymbol() == "O"
        one_hot = get_atom_explicit_valence_one_hot(atoms[3])
        assert one_hot == [0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 0.0]

    def test_get_bond_type_one_hot(self):
        bonds = self.mol.GetBonds()
        one_hot = get_bond_type_one_hot(bonds[0])
        # The C-N bond is a single bond
        assert bonds[0].GetBeginAtomIdx() == 0.0
        assert bonds[0].GetEndAtomIdx() == 1.0
        assert one_hot == [1.0, 0.0, 0.0, 0.0]

    def test_get_bond_is_in_same_ring_one_hot(self):
        bonds = self.mol.GetBonds()
        one_hot = get_bond_is_in_same_ring_one_hot(bonds[0])
        assert one_hot == [0.0]

        bonds = self.mol_benzene.GetBonds()
        one_hot = get_bond_is_in_same_ring_one_hot(bonds[0])
        assert one_hot == [1.0]

    def test_get_bond_is_conjugated_one_hot(self):
        bonds = self.mol.GetBonds()
        one_hot = get_bond_is_conjugated_one_hot(bonds[0])
        assert one_hot == [0.0]

        bonds = self.mol_benzene.GetBonds()
        one_hot = get_bond_is_conjugated_one_hot(bonds[0])
        assert one_hot == [1.0]

    def test_get_bond_stereo_one_hot(self):
        bonds = self.mol.GetBonds()
        one_hot = get_bond_stereo_one_hot(bonds[0])
        assert one_hot == [1.0, 0.0, 0.0, 0.0, 0.0]

    def test_get_bond_graph_distance_one_hot(self):
        from rdkit import Chem
        bonds = self.mol.GetBonds()
        dist_matrix = Chem.GetDistanceMatrix(self.mol)
        one_hot = get_bond_graph_distance_one_hot(bonds[0], dist_matrix)
        assert one_hot == [1.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0]



"""
Feature calculations.
"""
import inspect
import logging
import numpy as np
from typing import Any, Dict, Iterable, Optional, Tuple, Union, cast

from deepchem.utils import get_print_threshold
from deepchem.utils.typing import PymatgenStructure

logger = logging.getLogger(__name__)


class Featurizer(object):
    """Abstract class for calculating a set of features for a datapoint.

    This class is abstract and cannot be invoked directly. You'll
    likely only interact with this class if you're a developer. In
    that case, you might want to make a child class which
    implements the `_featurize` method for calculating features for
    a single datapoints if you'd like to make a featurizer for a
    new datatype.
    """

    def featurize(self,
                  datapoints: Iterable[Any],
                  log_every_n: int = 1000,
                  **kwargs) -> np.ndarray:
        """Calculate features for datapoints.

        Parameters
        ----------
        datapoints: Iterable[Any]
            A sequence of objects that you'd like to featurize. Subclassses of
            `Featurizer` should instantiate the `_featurize` method that featurizes
            objects in the sequence.
        log_every_n: int, default 1000
            Logs featurization progress every `log_every_n` steps.

        Returns
        -------
        np.ndarray
            A numpy array containing a featurized representation of `datapoints`.
        """
        datapoints = list(datapoints)
        features = []
        for i, point in enumerate(datapoints):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)
            try:
                features.append(self._featurize(point, **kwargs))
            except:
                logger.warning(
                    "Failed to featurize datapoint %d. Appending empty array")
                features.append(np.array([]))

        return np.asarray(features)

    def __call__(self, datapoints: Iterable[Any], **kwargs):
        """Calculate features for datapoints.

        `**kwargs` will get passed directly to `Featurizer.featurize`

        Parameters
        ----------
        datapoints: Iterable[Any]
            Any blob of data you like. Subclasss should instantiate this.
        """
        return self.featurize(datapoints, **kwargs)

    def _featurize(self, datapoint: Any, **kwargs):
        """Calculate features for a single datapoint.

        Parameters
        ----------
        datapoint: Any
            Any blob of data you like. Subclass should instantiate this.
        """
        raise NotImplementedError('Featurizer is not defined.')

    def __repr__(self) -> str:
        """Convert self to repr representation.

        Returns
        -------
        str
            The string represents the class.

        Examples
        --------
        >>> import deepchem as dc
        >>> dc.feat.CircularFingerprint(size=1024, radius=4)
        CircularFingerprint[radius=4, size=1024, chiral=False, bonds=True, features=False, sparse=False, smiles=False, is_counts_based=False]
        >>> dc.feat.CGCNNFeaturizer()
        CGCNNFeaturizer[radius=8.0, max_neighbors=12, step=0.2]
        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != 'self']
        args_info = ''
        for arg_name in args_names:
            value = self.__dict__[arg_name]
            # for str
            if isinstance(value, str):
                value = "'" + value + "'"
            # for list
            if isinstance(value, list):
                threshold = get_print_threshold()
                value = np.array2string(np.array(value), threshold=threshold)
            args_info += arg_name + '=' + str(value) + ', '
        return self.__class__.__name__ + '[' + args_info[:-2] + ']'

    def __str__(self) -> str:
        """Convert self to str representation.

        Returns
        -------
        str
            The string represents the class.

        Examples
        --------
        >>> import deepchem as dc
        >>> str(dc.feat.CircularFingerprint(size=1024, radius=4))
        'CircularFingerprint_radius_4_size_1024'
        >>> str(dc.feat.CGCNNFeaturizer())
        'CGCNNFeaturizer'
        """
        args_spec = inspect.getfullargspec(self.__init__)  # type: ignore
        args_names = [arg for arg in args_spec.args if arg != 'self']
        args_num = len(args_names)
        args_default_values = [None for _ in range(args_num)]
        if args_spec.defaults is not None:
            defaults = list(args_spec.defaults)
            args_default_values[-len(defaults):] = defaults

        override_args_info = ''
        for arg_name, default in zip(args_names, args_default_values):
            if arg_name in self.__dict__:
                arg_value = self.__dict__[arg_name]
                # validation
                # skip list
                if isinstance(arg_value, list):
                    continue
                if isinstance(arg_value, str):
                    # skip path string
                    if "\\/." in arg_value or "/" in arg_value or '.' in arg_value:
                        continue
                # main logic
                if default != arg_value:
                    override_args_info += '_' + arg_name + '_' + str(arg_value)
        return self.__class__.__name__ + override_args_info


class ComplexFeaturizer(Featurizer):
    """"
    Abstract class for calculating features for mol/protein complexes.
    """

    def featurize(self,
                  datapoints: Optional[Iterable[Tuple[str, str]]] = None,
                  log_every_n: int = 100,
                  **kwargs) -> np.ndarray:
        """
        Calculate features for mol/protein complexes.
        Parameters
        ----------
        datapoints: Iterable[Tuple[str, str]]
            List of filenames (PDB, SDF, etc.) for ligand molecules and proteins.
            Each element should be a tuple of the form (ligand_filename,
            protein_filename).

        Returns
        -------
        features: np.ndarray
            Array of features
        """

        if 'complexes' in kwargs:
            datapoints = kwargs.get("complexes")
            raise DeprecationWarning(
                'Complexes is being phased out as a parameter, please pass "datapoints" instead.'
            )
        if not isinstance(datapoints, Iterable):
            datapoints = [cast(Tuple[str, str], datapoints)]
        features, failures, successes = [], [], []
        for idx, point in enumerate(datapoints):
            if idx % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % idx)
            try:
                features.append(self._featurize(point, **kwargs))
                successes.append(idx)
            except:
                logger.warning(
                    "Failed to featurize datapoint %i. Appending empty array." %
                    idx)
                features.append(np.zeros(1))
                failures.append(idx)

        # Find a successful featurization
        try:
            i = np.argmax([f.shape[0] for f in features])
            dtype = features[i].dtype
            shape = features[i].shape
            dummy_array = np.zeros(shape, dtype=dtype)
        except AttributeError:
            dummy_array = features[successes[0]]

        # Replace failed featurizations with appropriate array
        for idx in failures:
            features[idx] = dummy_array

        return np.asarray(features)

    def _featurize(self, datapoint: Optional[Tuple[str, str]] = None, **kwargs):
        """
        Calculate features for single mol/protein complex.

        Parameters
        ----------
        complex: Tuple[str, str]
            Filenames for molecule and protein.
        """
        raise NotImplementedError('Featurizer is not defined.')


class MolecularFeaturizer(Featurizer):
    """Abstract class for calculating a set of features for a
molecule.

    The defining feature of a `MolecularFeaturizer` is that it
    uses SMILES strings and RDKit molecule objects to represent
    small molecules. All other featurizers which are subclasses of
    this class should plan to process input which comes as smiles
    strings or RDKit molecules.

    Child classes need to implement the _featurize method for
    calculating features for a single molecule.

    Note
    ----
    The subclasses of this class require RDKit to be installed.
    """

    def __init__(self, use_original_atoms_order=False):
        """
        Parameters
        ----------
        use_original_atoms_order: bool, default False
            Whether to use original atom ordering or canonical ordering (default)
        """
        self.use_original_atoms_order = use_original_atoms_order

    def featurize(self, datapoints, log_every_n=1000, **kwargs) -> np.ndarray:
        """Calculate features for molecules.

        Parameters
        ----------
        datapoints: rdkit.Chem.rdchem.Mol / SMILES string / iterable
            RDKit Mol, or SMILES string or iterable sequence of RDKit mols/SMILES
            strings.
        log_every_n: int, default 1000
            Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: np.ndarray
            A numpy array containing a featurized representation of `datapoints`.
        """
        try:
            from rdkit import Chem
            from rdkit.Chem import rdmolfiles
            from rdkit.Chem import rdmolops
            from rdkit.Chem.rdchem import Mol
        except ModuleNotFoundError:
            raise ImportError("This class requires RDKit to be installed.")

        if 'molecules' in kwargs:
            datapoints = kwargs.get("molecules")
            raise DeprecationWarning(
                'Molecules is being phased out as a parameter, please pass "datapoints" instead.'
            )

        # Special case handling of single molecule
        if isinstance(datapoints, str) or isinstance(datapoints, Mol):
            datapoints = [datapoints]
        else:
            # Convert iterables to list
            datapoints = list(datapoints)

        features: list = []
        for i, mol in enumerate(datapoints):
            if i % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % i)

            try:
                if isinstance(mol, str):
                    # condition if the original atom order is required
                    if hasattr(self, 'use_original_atoms_order'
                              ) and self.use_original_atoms_order:
                        # mol must be a RDKit Mol object, so parse a SMILES
                        mol = Chem.MolFromSmiles(mol)
                    else:
                        # mol must be a RDKit Mol object, so parse a SMILES
                        mol = Chem.MolFromSmiles(mol)
                        # SMILES is unique, so set a canonical order of atoms
                        new_order = rdmolfiles.CanonicalRankAtoms(mol)
                        mol = rdmolops.RenumberAtoms(mol, new_order)
                kwargs_per_datapoint = {}
                for key in kwargs.keys():
                    kwargs_per_datapoint[key] = kwargs[key][i]
                features.append(self._featurize(mol, **kwargs_per_datapoint))
            except Exception as e:
                if isinstance(mol, Chem.rdchem.Mol):
                    mol = Chem.MolToSmiles(mol)
                logger.warning(
                    "Failed to featurize datapoint %d, %s. Appending empty array",
                    i, mol)
                logger.warning("Exception message: {}".format(e))
                features.append(np.array([]))
        try:
            return np.asarray(features)
        except ValueError as e:
            logger.warning("Exception message: {}".format(e))
            return np.asarray(features, dtype=object)


class MaterialStructureFeaturizer(Featurizer):
    """
    Abstract class for calculating a set of features for an
    inorganic crystal structure.

    The defining feature of a `MaterialStructureFeaturizer` is that it
    operates on 3D crystal structures with periodic boundary conditions.
    Inorganic crystal structures are represented by Pymatgen structure
    objects. Featurizers for inorganic crystal structures that are subclasses of
    this class should plan to process input which comes as pymatgen
    structure objects.

    This class is abstract and cannot be invoked directly. You'll
    likely only interact with this class if you're a developer. Child
    classes need to implement the _featurize method for calculating
    features for a single crystal structure.

    Note
    ----
    Some subclasses of this class will require pymatgen and matminer to be
    installed.
    """

    def featurize(self,
                  datapoints: Optional[Iterable[Union[Dict[
                      str, Any], PymatgenStructure]]] = None,
                  log_every_n: int = 1000,
                  **kwargs) -> np.ndarray:
        """Calculate features for crystal structures.

        Parameters
        ----------
        datapoints: Iterable[Union[Dict, pymatgen.core.Structure]]
            Iterable sequence of pymatgen structure dictionaries
            or pymatgen.core.Structure. Please confirm the dictionary representations
            of pymatgen.core.Structure from https://pymatgen.org/pymatgen.core.structure.html.
        log_every_n: int, default 1000
            Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: np.ndarray
            A numpy array containing a featurized representation of
            `datapoints`.
        """
        try:
            from pymatgen.core import Structure
        except ModuleNotFoundError:
            raise ImportError("This class requires pymatgen to be installed.")

        if 'structures' in kwargs:
            datapoints = kwargs.get("structures")
            raise DeprecationWarning(
                'Structures is being phased out as a parameter, please pass "datapoints" instead.'
            )

        if not isinstance(datapoints, Iterable):
            datapoints = [
                cast(Union[Dict[str, Any], PymatgenStructure], datapoints)
            ]

        datapoints = list(datapoints)
        features = []
        for idx, structure in enumerate(datapoints):
            if idx % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % idx)
            try:
                if isinstance(structure, Dict):
                    structure = Structure.from_dict(structure)
                features.append(self._featurize(structure, **kwargs))
            except:
                logger.warning(
                    "Failed to featurize datapoint %i. Appending empty array" %
                    idx)
                features.append(np.array([]))

        return np.asarray(features)


class MaterialCompositionFeaturizer(Featurizer):
    """
    Abstract class for calculating a set of features for an
    inorganic crystal composition.

    The defining feature of a `MaterialCompositionFeaturizer` is that it
    operates on 3D crystal chemical compositions.
    Inorganic crystal compositions are represented by Pymatgen composition
    objects. Featurizers for inorganic crystal compositions that are
    subclasses of this class should plan to process input which comes as
    Pymatgen composition objects.

    This class is abstract and cannot be invoked directly. You'll
    likely only interact with this class if you're a developer. Child
    classes need to implement the _featurize method for calculating
    features for a single crystal composition.

    Note
    ----
    Some subclasses of this class will require pymatgen and matminer to be
    installed.
    """

    def featurize(self,
                  datapoints: Optional[Iterable[str]] = None,
                  log_every_n: int = 1000,
                  **kwargs) -> np.ndarray:
        """Calculate features for crystal compositions.

        Parameters
        ----------
        datapoints: Iterable[str]
            Iterable sequence of composition strings, e.g. "MoS2".
        log_every_n: int, default 1000
            Logging messages reported every `log_every_n` samples.

        Returns
        -------
        features: np.ndarray
            A numpy array containing a featurized representation of
            `compositions`.
        """
        try:
            from pymatgen.core import Composition
        except ModuleNotFoundError:
            raise ImportError("This class requires pymatgen to be installed.")

        if 'compositions' in kwargs and datapoints is None:
            datapoints = kwargs.get("compositions")
            raise DeprecationWarning(
                'Compositions is being phased out as a parameter, please pass "datapoints" instead.'
            )

        if not isinstance(datapoints, Iterable):
            datapoints = [cast(str, datapoints)]

        datapoints = list(datapoints)
        features = []
        for idx, composition in enumerate(datapoints):
            if idx % log_every_n == 0:
                logger.info("Featurizing datapoint %i" % idx)
            try:
                c = Composition(composition)
                features.append(self._featurize(c, **kwargs))
            except:
                logger.warning(
                    "Failed to featurize datapoint %i. Appending empty array" %
                    idx)
                features.append(np.array([]))

        return np.asarray(features)


class UserDefinedFeaturizer(Featurizer):
    """Directs usage of user-computed featurizations."""

    def __init__(self, feature_fields):
        """Creates user-defined-featurizer."""
        self.feature_fields = feature_fields


class DummyFeaturizer(Featurizer):
    """Class that implements a no-op featurization.
    This is useful when the raw dataset has to be used without featurizing the
    examples. The Molnet loader requires a featurizer input and such datasets
    can be used in their original form by passing the raw featurizer.

    Examples
    --------
    >>> import deepchem as dc
    >>> smi_map = [["N#C[S-].O=C(CBr)c1ccc(C(F)(F)F)cc1>CCO.[K+]", "N#CSCC(=O)c1ccc(C(F)(F)F)cc1"], ["C1COCCN1.FCC(Br)c1cccc(Br)n1>CCN(C(C)C)C(C)C.CN(C)C=O.O", "FCC(c1cccc(Br)n1)N1CCOCC1"]]
    >>> Featurizer = dc.feat.DummyFeaturizer()
    >>> smi_feat = Featurizer.featurize(smi_map)
    >>> smi_feat
    array([['N#C[S-].O=C(CBr)c1ccc(C(F)(F)F)cc1>CCO.[K+]',
            'N#CSCC(=O)c1ccc(C(F)(F)F)cc1'],
           ['C1COCCN1.FCC(Br)c1cccc(Br)n1>CCN(C(C)C)C(C)C.CN(C)C=O.O',
            'FCC(c1cccc(Br)n1)N1CCOCC1']], dtype='<U55')
    """

    def featurize(self,
                  datapoints: Iterable[Any],
                  log_every_n: int = 1000,
                  **kwargs) -> np.ndarray:
        """Passes through dataset, and returns the datapoint.

        Parameters
        ----
        datapoints: Iterable[Any]
            A sequence of objects that you'd like to featurize.

        Returns
        ----
        datapoints: np.ndarray
            A numpy array containing a featurized representation of
            the datapoints.
        """
        return np.asarray(datapoints)



# Requriments - transformers, tokenizers
# Right now, the Smiles Tokenizer uses an exiesting vocab file from rxnfp that is fairly comprehensive and from the USPTO dataset.
# The vocab may be expanded in the near future

import collections
import os
import re
import pkg_resources
from typing import List, Optional
from transformers import BertTokenizer
from logging import getLogger

logger = getLogger(__name__)
"""
SMI_REGEX_PATTERN: str
SMILES regex pattern for tokenization. Designed by Schwaller et. al.

References
----------
.. [1]  Philippe Schwaller, Teodoro Laino, Théophile Gaudin, Peter Bolgar, Christopher A. Hunter, Costas Bekas, and Alpha A. Lee
ACS Central Science 2019 5 (9): Molecular Transformer: A Model for Uncertainty-Calibrated Chemical Reaction Prediction
1572-1583 DOI: 10.1021/acscentsci.9b00576
"""

SMI_REGEX_PATTERN = r"""(\[[^\]]+]|Br?|Cl?|N|O|S|P|F|I|b|c|n|o|s|p|\(|\)|\.|=|#|-|\+|\\|\/|:|~|@|\?|>>?|\*|\$|\%[0-9]{2}|[0-9])"""

# add vocab_file dict
VOCAB_FILES_NAMES = {"vocab_file": "vocab.txt"}


def get_default_tokenizer():
    default_vocab_path = (pkg_resources.resource_filename(
        "deepchem", "feat/tests/vocab.txt"))
    return SmilesTokenizer(default_vocab_path)


class SmilesTokenizer(BertTokenizer):
    """
    Creates the SmilesTokenizer class. The tokenizer heavily inherits from the BertTokenizer
    implementation found in Huggingface's transformers library. It runs a WordPiece tokenization
    algorithm over SMILES strings using the tokenisation SMILES regex developed by Schwaller et. al.

    Please see https://github.com/huggingface/transformers
    and https://github.com/rxn4chemistry/rxnfp for more details.

    Examples
    --------
    >>> from deepchem.feat.smiles_tokenizer import SmilesTokenizer
    >>> current_dir = os.path.dirname(os.path.realpath(__file__))
    >>> vocab_path = os.path.join(current_dir, 'tests/data', 'vocab.txt')
    >>> tokenizer = SmilesTokenizer(vocab_path)
    >>> print(tokenizer.encode("CC(=O)OC1=CC=CC=C1C(=O)O"))
    [12, 16, 16, 17, 22, 19, 18, 19, 16, 20, 22, 16, 16, 22, 16, 16, 22, 16, 20, 16, 17, 22, 19, 18, 19, 13]


    References
    ----------
    .. [1] Schwaller, Philippe; Probst, Daniel; Vaucher, Alain C.; Nair, Vishnu H; Kreutter, David;
        Laino, Teodoro; et al. (2019): Mapping the Space of Chemical Reactions using Attention-Based Neural
        Networks. ChemRxiv. Preprint. https://doi.org/10.26434/chemrxiv.9897365.v3

    Note
    ----
    This class requires huggingface's transformers and tokenizers libraries to be installed.
    """
    vocab_files_names = VOCAB_FILES_NAMES

    def __init__(
        self,
        vocab_file: str = '',
        # unk_token="[UNK]",
        # sep_token="[SEP]",
        # pad_token="[PAD]",
        # cls_token="[CLS]",
        # mask_token="[MASK]",
        **kwargs):
        """Constructs a SmilesTokenizer.

        Parameters
        ----------
        vocab_file: str
            Path to a SMILES character per line vocabulary file.
            Default vocab file is found in deepchem/feat/tests/data/vocab.txt
        """

        super().__init__(vocab_file, **kwargs)

        if not os.path.isfile(vocab_file):
            raise ValueError(
                "Can't find a vocab file at path '{}'.".format(vocab_file))
        self.vocab = load_vocab(vocab_file)
        self.highest_unused_index = max([
            i for i, v in enumerate(self.vocab.keys())
            if v.startswith("[unused")
        ])
        self.ids_to_tokens = collections.OrderedDict([
            (ids, tok) for tok, ids in self.vocab.items()
        ])
        self.basic_tokenizer = BasicSmilesTokenizer()

    @property
    def vocab_size(self):
        return len(self.vocab)

    @property
    def vocab_list(self):
        return list(self.vocab.keys())

    def _tokenize(self, text: str, max_seq_length: int = 512, **kwargs):
        """Tokenize a string into a list of tokens.

        Parameters
        ----------
        text: str
            Input string sequence to be tokenized.
        """

        max_len_single_sentence = max_seq_length - 2
        split_tokens = [
            token for token in self.basic_tokenizer.tokenize(text)
            [:max_len_single_sentence]
        ]
        return split_tokens

    def _convert_token_to_id(self, token: str):
        """Converts a token (str/unicode) in an id using the vocab.

        Parameters
        ----------
        token: str
            String token from a larger sequence to be converted to a numerical id.
        """

        return self.vocab.get(token, self.vocab.get(self.unk_token))

    def _convert_id_to_token(self, index: int):
        """Converts an index (integer) in a token (string/unicode) using the vocab.

        Parameters
        ----------
        index: int
            Integer index to be converted back to a string-based token as part of a larger sequence.
        """

        return self.ids_to_tokens.get(index, self.unk_token)

    def convert_tokens_to_string(self, tokens: List[str]):
        """Converts a sequence of tokens (string) in a single string.

        Parameters
        ----------
        tokens: List[str]
            List of tokens for a given string sequence.

        Returns
        -------
        out_string: str
            Single string from combined tokens.
        """

        out_string: str = " ".join(tokens).replace(" ##", "").strip()
        return out_string

    def add_special_tokens_ids_single_sequence(self,
                                               token_ids: List[Optional[int]]):
        """Adds special tokens to the a sequence for sequence classification tasks.

        A BERT sequence has the following format: [CLS] X [SEP]

        Parameters
        ----------
        token_ids: list[int]
            list of tokenized input ids. Can be obtained using the encode or encode_plus methods.
        """

        return [self.cls_token_id] + token_ids + [self.sep_token_id]

    def add_special_tokens_single_sequence(self, tokens: List[str]):
        """Adds special tokens to the a sequence for sequence classification tasks.
        A BERT sequence has the following format: [CLS] X [SEP]

        Parameters
        ----------
        tokens: List[str]
            List of tokens for a given string sequence.
        """
        return [self.cls_token] + tokens + [self.sep_token]

    def add_special_tokens_ids_sequence_pair(
            self, token_ids_0: List[Optional[int]],
            token_ids_1: List[Optional[int]]) -> List[Optional[int]]:
        """Adds special tokens to a sequence pair for sequence classification tasks.
        A BERT sequence pair has the following format: [CLS] A [SEP] B [SEP]

        Parameters
        ----------
        token_ids_0: List[int]
            List of ids for the first string sequence in the sequence pair (A).
        token_ids_1: List[int]
            List of tokens for the second string sequence in the sequence pair (B).
        """

        sep = [self.sep_token_id]
        cls = [self.cls_token_id]

        return cls + token_ids_0 + sep + token_ids_1 + sep

    def add_padding_tokens(self,
                           token_ids: List[Optional[int]],
                           length: int,
                           right: bool = True) -> List[Optional[int]]:
        """Adds padding tokens to return a sequence of length max_length.
        By default padding tokens are added to the right of the sequence.

        Parameters
        ----------
        token_ids: list[optional[int]]
            list of tokenized input ids. Can be obtained using the encode or encode_plus methods.
        length: int
            TODO
        right: bool, default True
            TODO

        Returns
        -------
        List[int]
            TODO
        """
        padding = [self.pad_token_id] * (length - len(token_ids))

        if right:
            return token_ids + padding
        else:
            return padding + token_ids

    def save_vocabulary(
        self,
        save_directory: str,
        filename_prefix: Optional[str] = None
    ):  # -> Tuple[str]: doctest issue raised with this return type annotation
        """Save the tokenizer vocabulary to a file.

        Parameters
        ----------
        vocab_path: obj: str
            The directory in which to save the SMILES character per line vocabulary file.
            Default vocab file is found in deepchem/feat/tests/data/vocab.txt

        Returns
        -------
        vocab_file: Tuple
            Paths to the files saved.
            typle with string to a SMILES character per line vocabulary file.
            Default vocab file is found in deepchem/feat/tests/data/vocab.txt
        """
        index = 0
        if os.path.isdir(save_directory):
            vocab_file = os.path.join(save_directory,
                                      VOCAB_FILES_NAMES["vocab_file"])
        else:
            vocab_file = save_directory
        with open(vocab_file, "w", encoding="utf-8") as writer:
            for token, token_index in sorted(self.vocab.items(),
                                             key=lambda kv: kv[1]):
                if index != token_index:
                    logger.warning(
                        "Saving vocabulary to {}: vocabulary indices are not consecutive."
                        " Please check that the vocabulary is not corrupted!".
                        format(vocab_file))
                    index = token_index
                writer.write(token + "\n")
                index += 1
        return (vocab_file,)


class BasicSmilesTokenizer(object):
    """
    Run basic SMILES tokenization using a regex pattern developed by Schwaller et. al.
    This tokenizer is to be used when a tokenizer that does not require the transformers library by HuggingFace is required.

    Examples
    --------
    >>> from deepchem.feat.smiles_tokenizer import BasicSmilesTokenizer
    >>> tokenizer = BasicSmilesTokenizer()
    >>> print(tokenizer.tokenize("CC(=O)OC1=CC=CC=C1C(=O)O"))
    ['C', 'C', '(', '=', 'O', ')', 'O', 'C', '1', '=', 'C', 'C', '=', 'C', 'C', '=', 'C', '1', 'C', '(', '=', 'O', ')', 'O']


    References
    ----------
    .. [1] Philippe Schwaller, Teodoro Laino, Théophile Gaudin, Peter Bolgar, Christopher A. Hunter, Costas Bekas, and Alpha A. Lee
        ACS Central Science 2019 5 (9): Molecular Transformer: A Model for Uncertainty-Calibrated Chemical Reaction Prediction
        1572-1583 DOI: 10.1021/acscentsci.9b00576
    """

    def __init__(self, regex_pattern: str = SMI_REGEX_PATTERN):
        """Constructs a BasicSMILESTokenizer.

        Parameters
        ----------
        regex: string
            SMILES token regex
        """
        self.regex_pattern = regex_pattern
        self.regex = re.compile(self.regex_pattern)

    def tokenize(self, text):
        """Basic Tokenization of a SMILES.
        """
        tokens = [token for token in self.regex.findall(text)]
        return tokens


def load_vocab(vocab_file):
    """Loads a vocabulary file into a dictionary."""
    vocab = collections.OrderedDict()
    with open(vocab_file, "r", encoding="utf-8") as reader:
        tokens = reader.readlines()
    for index, token in enumerate(tokens):
        token = token.rstrip("\n")
        vocab[token] = index
    return vocab



import numpy as np

from deepchem.utils.typing import RDKitMol
from deepchem.utils.data_utils import pad_array
from deepchem.feat.base_classes import MolecularFeaturizer
from deepchem.feat.molecule_featurizers.atomic_coordinates import AtomicCoordinates


class BPSymmetryFunctionInput(MolecularFeaturizer):
    """Calculate symmetry function for each atom in the molecules

    This method is described in [1]_.

    Examples
    --------
    >>> import deepchem as dc
    >>> smiles = ['C1C=CC=CC=1']
    >>> featurizer = dc.feat.BPSymmetryFunctionInput(max_atoms=10)
    >>> features = featurizer.featurize(smiles)
    >>> type(features[0])
    <class 'numpy.ndarray'>
    >>> features[0].shape  # (max_atoms, 4)
    (10, 4)

    References
    ----------
    .. [1] Behler, Jörg, and Michele Parrinello. "Generalized neural-network
        representation of high-dimensional potential-energy surfaces." Physical
        review letters 98.14 (2007): 146401.

    Note
    ----
    This class requires RDKit to be installed.

    """

    def __init__(self, max_atoms: int):
        """Initialize this featurizer.

        Parameters
        ----------
        max_atoms: int
            The maximum number of atoms expected for molecules this featurizer will
            process.

        """
        self.max_atoms = max_atoms
        self.coordfeat = AtomicCoordinates(use_bohr=True)

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> np.ndarray:
        """Calculate symmetry function.

        Parameters
        ----------
        datapoint: rdkit.Chem.rdchem.Mol
            RDKit Mol object

        Returns
        -------
        np.ndarray
            A numpy array of symmetry function. The shape is `(max_atoms, 4)`.

        """
        if 'mol' in kwargs:
            datapoint = kwargs.get("mol")
            raise DeprecationWarning(
                'Mol is being phased out as a parameter, please pass "datapoint" instead.'
            )
        coordinates = self.coordfeat._featurize(datapoint)
        atom_numbers = np.array(
            [atom.GetAtomicNum() for atom in datapoint.GetAtoms()])
        atom_numbers = np.expand_dims(atom_numbers, axis=1)
        assert atom_numbers.shape[0] == coordinates.shape[0]
        features = np.concatenate([atom_numbers, coordinates], axis=1)
        return pad_array(features, (self.max_atoms, 4))



import numpy as np
import time
import logging
from collections.abc import Sequence as SequenceCollection

from deepchem.data import Dataset, NumpyDataset
from deepchem.metrics import Metric
from deepchem.models.models import Model
from deepchem.models.losses import Loss
from deepchem.models.optimizers import Optimizer, Adam
from deepchem.utils.evaluate import GeneratorEvaluator
from deepchem.trans.transformers import Transformer, undo_transforms

from typing import Any, Callable, Iterable, List, Optional, Tuple, Union, Sequence
from deepchem.utils.typing import LossFn, OneOrMany

# JAX depend
import jax.numpy as jnp
import jax
import haiku as hk
import optax

import warnings

logger = logging.getLogger(__name__)


def create_default_eval_fn(forward_fn, params):
    """
    Calls the function to evaluate the model
    """

    @jax.jit
    def eval_model(batch, rng=None):
        predict = forward_fn(params, rng, batch)

        return predict

    return eval_model


def create_default_update_fn(optimizer, model_loss):
    """
    This function calls the update function, to implement the backpropogation
    """

    @jax.jit
    def update(params, opt_state, batch, target, weights,
               rng) -> Tuple[hk.Params, optax.OptState, jnp.ndarray]:
        batch_loss, grads = jax.value_and_grad(model_loss)(params, batch,
                                                           target, weights, rng)
        updates, opt_state = optimizer.update(grads, opt_state)
        new_params = optax.apply_updates(params, updates)
        return new_params, opt_state, batch_loss

    return update


def create_default_gradient_fn(forward_fn, loss_outputs, loss_fn):
    """
    This function calls the gradient function, to implement the backpropogation
    """

    @jax.jit
    def model_loss(params, batch, target, weights, rng):
        predict = forward_fn(params, rng, batch)
        if loss_outputs is not None:
            predict = [predict[i] for i in loss_outputs]
        return loss_fn(predict, target, weights)

    return model_loss


class JaxModel(Model):
    """This is a DeepChem model implemented by a Jax Model
    Here is a simple example of that uses JaxModel to train a
    Haiku (JAX Neural Network Library) based model on deepchem
    dataset.

    >>>
    >> def forward_model(x):
    >>   net = hk.nets.MLP([512, 256, 128, 1])
    >>   return net(x)
    >> def rms_loss(pred, tar, w):
    >>   return jnp.mean(optax.l2_loss(pred, tar))
    >> params_init, forward_fn = hk.transform(forward_model)
    >> rng = jax.random.PRNGKey(500)
    >> inputs, _, _, _ = next(iter(dataset.iterbatches(batch_size=256)))
    >> params = params_init(rng, inputs)
    >> j_m = JaxModel(forward_fn, params, rms_loss, 256, 0.001, 100)
    >> j_m.fit(train_dataset)

    All optimizations will be done using the optax library.
    """

    def __init__(self,
                 forward_fn: hk.State,
                 params: hk.Params,
                 loss: Optional[Union[Loss, LossFn]],
                 output_types: Optional[List[str]] = None,
                 batch_size: int = 100,
                 learning_rate: float = 0.001,
                 optimizer: Optional[Union[optax.GradientTransformation,
                                           Optimizer]] = None,
                 grad_fn: Callable = create_default_gradient_fn,
                 update_fn: Callable = create_default_update_fn,
                 eval_fn: Callable = create_default_eval_fn,
                 rng=jax.random.PRNGKey(1),
                 log_frequency: int = 100,
                 **kwargs):
        """
        Create a new JaxModel

        Parameters
        ----------
        model: hk.State or Function
            Any Jax based model that has a `apply` method for computing the network. Currently
            only haiku models are supported.
        params: hk.Params
            The parameter of the Jax based networks
        loss: dc.models.losses.Loss or function
            a Loss or function defining how to compute the training loss for each
            batch, as described above
        output_types: list of strings, optional (default None)
            the type of each output from the model, as described above
        batch_size: int, optional (default 100)
            default batch size for training and evaluating
        learning_rate: float or LearningRateSchedule, optional (default 0.001)
            the learning rate to use for fitting.  If optimizer is specified, this is
            ignored.
        optimizer: optax object
            For the time being, it is optax object
        rng: jax.random.PRNGKey, optional (default 1)
            A default global PRNG key to use for drawing random numbers.
        log_frequency: int, optional (default 100)
            The frequency at which to log data. Data is logged using
            `logging` by default.


        Miscellanous Parameters Yet To Add
        ----------------------------------
        model_dir: str, optional (default None)
            Will be added along with the save & load method
        tensorboard: bool, optional (default False)
            whether to log progress to TensorBoard during training
        wandb: bool, optional (default False)
            whether to log progress to Weights & Biases during training


        Work in Progress
        ----------------
        [1] Integrate the optax losses, optimizers, schedulers with Deepchem
        [2] Support for saving & loading the model.
        """
        super(JaxModel, self).__init__(model=(forward_fn, params), **kwargs)
        warnings.warn(
            'JaxModel is still in active development and all features may not yet be implemented'
        )
        self._loss_fn = loss  # lambda pred, tar: jnp.mean(optax.l2_loss(pred, tar))
        self.batch_size = batch_size
        self.learning_rate = learning_rate
        if optimizer is None:
            optimizer = Adam(1e-3)

        if not isinstance(optimizer, optax.GradientTransformation):
            self.optimizer = optimizer._create_jax_optimizer()
        else:
            self.optimizer = optimizer
        self.forward_fn = forward_fn
        self.params = params
        self._built = False
        self.log_frequency = log_frequency
        self.rng = rng
        self._create_gradient_fn = grad_fn
        self._create_update_fn = update_fn
        self._create_eval_fn = eval_fn

        if output_types is None:
            self._prediction_outputs = None
            self._loss_outputs = None
            self._variance_outputs = None
            self._other_outputs = None
        else:
            self._prediction_outputs = []
            self._loss_outputs = []
            self._variance_outputs = []
            self._other_outputs = []
            for i, type in enumerate(output_types):
                if type == 'prediction':
                    self._prediction_outputs.append(i)
                elif type == 'loss':
                    self._loss_outputs.append(i)
                elif type == 'variance':
                    self._variance_outputs.append(i)
                else:
                    self._other_outputs.append(i)
            if len(self._loss_outputs) == 0:
                self._loss_outputs = self._prediction_outputs

    def _ensure_built(self):
        """The first time this is called, create internal data structures.
        Work in Progress
        ----------------
        [1] Integerate the optax losses, optimizers, schedulers with Deepchem
        """
        if self._built:
            return

        self._built = True
        self._global_step = 0
        self.opt_state = self.optimizer.init(self.params)

    def fit(self,
            dataset: Dataset,
            nb_epochs: int = 10,
            deterministic: bool = False,
            loss: Optional[Union[Loss, LossFn]] = None,
            callbacks: Union[Callable, List[Callable]] = [],
            all_losses: Optional[List[float]] = None) -> float:
        """Train this model on a dataset.
        Parameters
        ----------
        dataset: Dataset
            the Dataset to train on
        nb_epoch: int
            the number of epochs to train for
        deterministic: bool
            if True, the samples are processed in order.  If False, a different random
            order is used for each epoch.
        loss: function
            a function of the form f(outputs, labels, weights) that computes the loss
            for each batch.  If None (the default), the model's standard loss function
            is used.
        callbacks: function or list of functions
            one or more functions of the form f(model, step) that will be invoked after
            every step.  This can be used to perform validation, logging, etc.
        all_losses: Optional[List[float]], optional (default None)
            If specified, all logged losses are appended into this list. Note that
            you can call `fit()` repeatedly with the same list and losses will
            continue to be appended.
        Returns
        -------
        The average loss over the most recent checkpoint interval
        Miscellanous Parameters Yet To Add
        ----------------------------------
        max_checkpoints_to_keep: int
            the maximum number of checkpoints to keep.  Older checkpoints are discarded.
        checkpoint_interval: int
            the frequency at which to write checkpoints, measured in training steps.
            Set this to 0 to disable automatic checkpointing.
        restore: bool
            if True, restore the model from the most recent checkpoint and continue training
            from there.  If False, retrain the model from scratch.
        variables: list of hk.Variable
            the variables to train.  If None (the default), all trainable variables in
            the model are used.
        Work in Progress
        ----------------
        [1] Integerate the optax losses, optimizers, schedulers with Deepchem
        [2] Support for saving & loading the model.
        [3] Adding support for output types (choosing only self._loss_outputs)
        """
        return self.fit_generator(
            self.default_generator(dataset,
                                   epochs=nb_epochs,
                                   deterministic=deterministic), loss,
            callbacks, all_losses)

    def fit_generator(self,
                      generator: Iterable[Tuple[Any, Any, Any]],
                      loss: Optional[Union[Loss, LossFn]] = None,
                      callbacks: Union[Callable, List[Callable]] = [],
                      all_losses: Optional[List[float]] = None) -> float:
        if not isinstance(callbacks, SequenceCollection):
            callbacks = [callbacks]
        self._ensure_built()
        avg_loss = 0.0
        last_avg_loss = 0.0
        averaged_batches = 0
        if loss is None:
            loss = self._loss_fn
        model_loss_fn = self._create_gradient_fn(self.forward_fn,
                                                 self._loss_outputs, loss)
        grad_update = self._create_update_fn(self.optimizer, model_loss_fn)

        params, opt_state = self._get_trainable_params()
        rng = self.rng
        time1 = time.time()

        # Main training loop

        for batch in generator:
            inputs, labels, weights = self._prepare_batch(batch)

            if isinstance(inputs, list) and len(inputs) == 1:
                inputs = inputs[0]

            if isinstance(labels, list) and len(labels) == 1:
                labels = labels[0]

            if isinstance(weights, list) and len(weights) == 1:
                weights = weights[0]

            params, opt_state, batch_loss = grad_update(params,
                                                        opt_state,
                                                        inputs,
                                                        labels,
                                                        weights,
                                                        rng=rng)
            rng, _ = jax.random.split(rng)
            avg_loss += jax.device_get(batch_loss)
            self._global_step += 1
            current_step = self._global_step
            averaged_batches += 1
            should_log = (current_step % self.log_frequency == 0)

            if should_log:
                avg_loss = float(avg_loss) / averaged_batches
                logger.info('Ending global_step %d: Average loss %g' %
                            (current_step, avg_loss))
                if all_losses is not None:
                    all_losses.append(avg_loss)
                # Capture the last avg_loss in case of return since we're resetting to 0 now
                last_avg_loss = avg_loss
                avg_loss = 0.0
                averaged_batches = 0
            for c in callbacks:
                c(self, current_step)

        # Report final results.
        if averaged_batches > 0:
            avg_loss = float(avg_loss) / averaged_batches
            logger.info('Ending global_step %d: Average loss %g' %
                        (current_step, avg_loss))
            if all_losses is not None:
                all_losses.append(avg_loss)
            last_avg_loss = avg_loss

        time2 = time.time()
        logger.info("TIMING: model fitting took %0.3f s" % (time2 - time1))
        self._set_trainable_params(params, opt_state)
        return last_avg_loss

    def _predict(self, generator: Iterable[Tuple[Any, Any, Any]],
                 transformers: List[Transformer], uncertainty: bool,
                 other_output_types: Optional[OneOrMany[str]]):
        """
        Predict outputs for data provided by a generator.
        This is the private implementation of prediction.  Do not
        call it directly. Instead call one of the public prediction
        methods.

        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        transformers: List[dc.trans.Transformers]
            Transformers that the input data has been transformed by.  The output
            is passed through these transformers to undo the transformations.
        uncertainty: bool
            specifies whether this is being called as part of estimating uncertainty.
            If True, it sets the training flag so that dropout will be enabled, and
            returns the values of the uncertainty outputs.
        other_output_types: list, optional
            Provides a list of other output_types (strings) to predict from model.

        Returns
        -------
        A NumpyArray if the model produces a single output, or a list of arrays otherwise.
        """
        results: Optional[List[List[np.ndarray]]] = None
        variances: Optional[List[List[np.ndarray]]] = None
        if uncertainty and (other_output_types is not None):
            raise ValueError(
                'This model cannot compute uncertainties and other output types simultaneously. Please invoke one at a time.'
            )
        if uncertainty:
            if self._variance_outputs is None or len(
                    self._variance_outputs) == 0:
                raise ValueError('This model cannot compute uncertainties')
            if len(self._variance_outputs) != len(self._prediction_outputs):
                raise ValueError(
                    'The number of variances must exactly match the number of outputs'
                )
        if other_output_types:
            if self._other_outputs is None or len(self._other_outputs) == 0:
                raise ValueError(
                    'This model cannot compute other outputs since no other output_types were specified.'
                )
        self._ensure_built()
        eval_fn = self._create_eval_fn(self.forward_fn, self.params)
        rng = self.rng

        for batch in generator:
            inputs, _, _ = self._prepare_batch(batch)

            if isinstance(inputs, list) and len(inputs) == 1:
                inputs = inputs[0]

            output_values = eval_fn(inputs, rng)
            if isinstance(output_values, jnp.ndarray):
                output_values = [output_values]
            output_values = [jax.device_get(t) for t in output_values]

            # Apply tranformers and record results.
            if uncertainty:
                var = [output_values[i] for i in self._variance_outputs]
                if variances is None:
                    variances = [var]
                else:
                    for i, t in enumerate(var):
                        variances[i].append(t)

            access_values = []
            if other_output_types:
                access_values += self._other_outputs
            elif self._prediction_outputs is not None:
                access_values += self._prediction_outputs

            if len(access_values) > 0:
                output_values = [output_values[i] for i in access_values]

            if len(transformers) > 0:
                if len(output_values) > 1:
                    raise ValueError(
                        "predict() does not support Transformers for models with multiple outputs."
                    )
                elif len(output_values) == 1:
                    output_values = [
                        undo_transforms(output_values[0], transformers)
                    ]

            if results is None:
                results = [[] for i in range(len(output_values))]
            for i, t in enumerate(output_values):
                results[i].append(t)

        # Concatenate arrays to create the final results.
        final_results = []
        final_variances = []
        if results is not None:
            for r in results:
                final_results.append(np.concatenate(r, axis=0))
        if uncertainty and variances is not None:
            for v in variances:
                final_variances.append(np.concatenate(v, axis=0))
            return zip(final_results, final_variances)
        if len(final_results) == 1:
            return final_results[0]
        else:
            return final_results

    def predict_on_generator(
            self,
            generator: Iterable[Tuple[Any, Any, Any]],
            transformers: List[Transformer] = [],
            output_types: Optional[OneOrMany[str]] = None
    ) -> OneOrMany[np.ndarray]:
        """
        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        transformers: List[dc.trans.Transformers]
            Transformers that the input data has been transformed by.  The output
            is passed through these transformers to undo the transformations.
        output_types: String or list of Strings
            If specified, all outputs of this type will be retrieved
            from the model. If output_types is specified, outputs must
            be None.

        Returns
        -------
            a NumPy array of the model produces a single output, or a list of arrays
            if it produces multiple outputs
        """
        return self._predict(generator, transformers, False, output_types)

    def predict_on_batch(
            self,
            X: np.typing.ArrayLike,
            transformers: List[Transformer] = []) -> OneOrMany[np.ndarray]:
        """Generates predictions for input samples, processing samples in a batch.
        Parameters
        ----------
        X: ndarray
            the input data, as a Numpy array.
        transformers: List[dc.trans.Transformers]
            Transformers that the input data has been transformed by.  The output
            is passed through these transformers to undo the transformations.

        Returns
        -------
        a NumPy array of the model produces a single output, or a list of arrays
        if it produces multiple outputs
        """
        dataset = NumpyDataset(X=X, y=None)
        return self.predict(dataset, transformers)

    def predict_uncertainty_on_batch(
            self,
            X: Sequence,
            masks: int = 50) -> OneOrMany[Tuple[np.ndarray, np.ndarray]]:
        raise NotImplementedError(
            'Predicting uncertainity on batch is not supported currently for JAX models'
        )

    def predict(
            self,
            dataset: Dataset,
            transformers: List[Transformer] = [],
            output_types: Optional[List[str]] = None) -> OneOrMany[np.ndarray]:
        """
        Uses self to make predictions on provided Dataset object.

        Parameters
        ----------
        dataset: dc.data.Dataset
            Dataset to make prediction on
        transformers: List[dc.trans.Transformers]
            Transformers that the input data has been transformed by.  The output
            is passed through these transformers to undo the transformations.
        output_types: String or list of Strings
            If specified, all outputs of this type will be retrieved
            from the model. If output_types is specified, outputs must
            be None.

        Returns
        -------
        a NumPy array of the model produces a single output, or a list of arrays
        if it produces multiple outputs
        """
        generator = self.default_generator(dataset,
                                           mode='predict',
                                           pad_batches=False)
        return self.predict_on_generator(generator,
                                         transformers=transformers,
                                         output_types=output_types)

    def get_global_step(self) -> int:
        """Get the number of steps of fitting that have been performed."""
        return self._global_step

    def predict_embedding(self, dataset: Dataset) -> OneOrMany[np.ndarray]:
        raise NotImplementedError(
            'Predicting embedding is not supported currently for JAX models')

    # def predict_uncertainty(self, dataset: Dataset, masks: int = 50
    #                        ) -> OneOrMany[Tuple[np.ndarray, np.ndarray]]:
    #   """
    #   Predict the model's outputs, along with the uncertainty in each one.
    #   The uncertainty is computed as described in https://arxiv.org/abs/1703.04977.
    #   It involves repeating the prediction many times with different dropout masks.
    #   The prediction is computed as the average over all the predictions.  The
    #   uncertainty includes both the variation among the predicted values (epistemic
    #   uncertainty) and the model's own estimates for how well it fits the data
    #   (aleatoric uncertainty).  Not all models support uncertainty prediction.
    #   Parameters
    #   ----------
    #   dataset: dc.data.Dataset
    #     Dataset to make prediction on
    #   masks: int
    #     the number of dropout masks to average over
    #   Returns
    #   -------
    #   for each output, a tuple (y_pred, y_std) where y_pred is the predicted
    #   value of the output, and each element of y_std estimates the standard
    #   deviation of the corresponding element of y_pred
    #   """
    #   sum_pred: List[np.ndarray] = []
    #   sum_sq_pred: List[np.ndarray] = []
    #   sum_var: List[np.ndarray] = []
    #   for i in range(masks):
    #     generator = self.default_generator(
    #         dataset, mode='uncertainty', pad_batches=False)
    #     results = self._predict(generator, [], True, None)
    #     if len(sum_pred) == 0:
    #       for p, v in results:
    #         sum_pred.append(p)
    #         sum_sq_pred.append(p * p)
    #         sum_var.append(v)
    #     else:
    #       for j, (p, v) in enumerate(results):
    #         sum_pred[j] += p
    #         sum_sq_pred[j] += p * p
    #         sum_var[j] += v
    #   output = []
    #   std = []
    #   for i in range(len(sum_pred)):
    #     p = sum_pred[i] / masks
    #     output.append(p)
    #     std.append(np.sqrt(sum_sq_pred[i] / masks - p * p + sum_var[i] / masks))
    #   if len(output) == 1:
    #     return (output[0], std[0])
    #   else:
    #     return list(zip(output, std))

    def evaluate_generator(self,
                           generator: Iterable[Tuple[Any, Any, Any]],
                           metrics: List[Metric],
                           transformers: List[Transformer] = [],
                           per_task_metrics: bool = False):
        """Evaluate the performance of this model on the data produced by a generator.
        Parameters
        ----------
        generator: generator
            this should generate batches, each represented as a tuple of the form
            (inputs, labels, weights).
        metric: list of deepchem.metrics.Metric
            Evaluation metric
        transformers: List[dc.trans.Transformers]
            Transformers that the input data has been transformed by.  The output
            is passed through these transformers to undo the transformations.
        per_task_metrics: bool
            If True, return per-task scores.
        Returns
        -------
        dict
            Maps tasks to scores under metric.
        """
        evaluator = GeneratorEvaluator(self, generator, transformers)
        return evaluator.compute_model_performance(metrics, per_task_metrics)

    def _get_trainable_params(self):
        """
        Will be used to seperate freezing parameters while transfer learning
        """
        return self.params, self.opt_state

    def _set_trainable_params(self, params: hk.Params,
                              opt_state: optax.OptState):
        """
        A functional approach to setting the final parameters after training
        """
        self.params = params
        self.opt_state = opt_state

    def _prepare_batch(self, batch):
        inputs, labels, weights = batch
        inputs = [
            x.astype(np.float32) if x.dtype == np.float64 else x for x in inputs
        ]
        if labels is not None:
            labels = [
                x.astype(np.float32) if x.dtype == np.float64 else x
                for x in labels
            ]
        else:
            labels = []

        if weights is not None:
            weights = [
                x.astype(np.float32) if x.dtype == np.float64 else x
                for x in weights
            ]
        else:
            weights = []

        return (inputs, labels, weights)

    def default_generator(
            self,
            dataset: Dataset,
            epochs: int = 1,
            mode: str = 'fit',
            deterministic: bool = True,
            pad_batches: bool = True) -> Iterable[Tuple[List, List, List]]:
        """Create a generator that iterates batches for a dataset.
        Subclasses may override this method to customize how model inputs are
        generated from the data.
        Parameters
        ----------
        dataset: Dataset
            the data to iterate
        epochs: int
            the number of times to iterate over the full dataset
        mode: str
            allowed values are 'fit' (called during training), 'predict' (called
            during prediction), and 'uncertainty' (called during uncertainty
            prediction)
        deterministic: bool
            whether to iterate over the dataset in order, or randomly shuffle the
            data for each epoch
        pad_batches: bool
            whether to pad each batch up to this model's preferred batch size
        Returns
        -------
        a generator that iterates batches, each represented as a tuple of lists:
        ([inputs], [outputs], [weights])
        """

        for epoch in range(epochs):
            for (X_b, y_b, w_b,
                 _) in dataset.iterbatches(batch_size=self.batch_size,
                                           deterministic=deterministic,
                                           pad_batches=pad_batches):
                yield ([X_b], [y_b], [w_b])



import numpy as np

try:
    import torch
except ImportError:
    print("This module requires PyTorch to be installed.")


def fourier_encode_dist(x, num_encodings=4, include_self=True):
    """
    Fourier encode the input tensor `x` based on the specified number of encodings.

    This function applies a Fourier encoding to the input tensor `x` by dividing
    it by a range of scales (2^i for i in range(num_encodings)) and then
    concatenating the sine and cosine of the scaled values. Optionally, the
    original input tensor can be included in the output.

    Parameters
    ----------
    x : torch.Tensor
        Input tensor to be Fourier encoded.
    num_encodings : int, optional, default=4
        Number of Fourier encodings to apply.
    include_self : bool, optional, default=True
        Whether to include the original input tensor in the output.

    Returns
    -------
    torch.Tensor
        Fourier encoded tensor.

    Examples
    --------
    >>> import torch
    >>> x = torch.tensor([1.0, 2.0, 3.0])
    >>> encoded_x = fourier_encode_dist(x, num_encodings=4, include_self=True)
    """
    x = x.unsqueeze(-1)
    device, dtype, orig_x = x.device, x.dtype, x
    scales = 2**torch.arange(num_encodings, device=device, dtype=dtype)
    x = x / scales
    x = torch.cat([x.sin(), x.cos()], dim=-1)
    x = torch.cat((x, orig_x), dim=-1) if include_self else x
    return x.squeeze()


EPS = 1e-5


def aggregate_mean(h, **kwargs):
    """
    Compute the mean of the input tensor along the second to last dimension.

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.

    Returns
    -------
    torch.Tensor
        Mean of the input tensor along the second to last dimension.
    """
    return torch.mean(h, dim=-2)


def aggregate_max(h, **kwargs):
    """
    Compute the max of the input tensor along the second to last dimension.

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.

    Returns
    -------
    torch.Tensor
        Max of the input tensor along the second to last dimension.
    """
    return torch.max(h, dim=-2)[0]


def aggregate_min(h, **kwargs):
    """
    Compute the min of the input tensor along the second to last dimension.

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.
    **kwargs
        Additional keyword arguments.

    Returns
    -------
    torch.Tensor
        Min of the input tensor along the second to last dimension.
    """
    return torch.min(h, dim=-2)[0]


def aggregate_std(h, **kwargs):
    """
    Compute the standard deviation of the input tensor along the second to last dimension.

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.

    Returns
    -------
    torch.Tensor
        Standard deviation of the input tensor along the second to last dimension.
    """
    return torch.sqrt(aggregate_var(h) + EPS)


def aggregate_var(h, **kwargs):
    """
    Compute the variance of the input tensor along the second to last dimension.

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.

    Returns
    -------
    torch.Tensor
        Variance of the input tensor along the second to last dimension.
    """
    h_mean_squares = torch.mean(h * h, dim=-2)
    h_mean = torch.mean(h, dim=-2)
    var = torch.relu(h_mean_squares - h_mean * h_mean)
    return var


def aggregate_moment(h, n=3, **kwargs):
    """
    Compute the nth moment of the input tensor along the second to last dimension.

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.
    n : int, optional, default=3
        The order of the moment to compute.

    Returns
    -------
    torch.Tensor
        Nth moment of the input tensor along the second to last dimension.
    """
    # for each node (E[(X-E[X])^n])^{1/n}
    # EPS is added to the absolute value of expectation before taking the nth root for stability
    h_mean = torch.mean(h, dim=-2, keepdim=True)
    h_n = torch.mean(torch.pow(h - h_mean, n), dim=-2)
    rooted_h_n = torch.sign(h_n) * torch.pow(torch.abs(h_n) + EPS, 1.0 / n)
    return rooted_h_n


def aggregate_sum(h, **kwargs):
    """
    Compute the sum of the input tensor along the second to last dimension.

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.

    Returns
    -------
    torch.Tensor
        Sum of the input tensor along the second to last dimension.
    """
    return torch.sum(h, dim=-2)


# each scaler is a function that takes as input X (B x N x Din), adj (B x N x N) and
# avg_d (dictionary containing averages over training set) and returns X_scaled (B x N x Din) as output
def scale_identity(h, D=None, avg_d=None):
    """
    Identity scaling function.

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.
    D : torch.Tensor, optional
        Degree tensor.
    avg_d : dict, optional
        Dictionary containing averages over the training set.

    Returns
    -------
    torch.Tensor
        Scaled input tensor.
    """
    return h


def scale_amplification(h, D, avg_d):
    """
    Amplification scaling function. log(D + 1) / d * h where d is the average of the ``log(D + 1)`` in the training set

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.
    D : torch.Tensor
        Degree tensor.
    avg_d : dict
        Dictionary containing averages over the training set.

    Returns
    -------
    torch.Tensor
        Scaled input tensor.
    """
    return h * (np.log(D + 1) / avg_d["log"])


def scale_attenuation(h, D, avg_d):
    """
    Attenuation scaling function. (log(D + 1))^-1 / d * X where d is the average of the ``log(D + 1))^-1`` in the training set

    Parameters
    ----------
    h : torch.Tensor
        Input tensor.
    D : torch.Tensor
        Degree tensor.
    avg_d : dict
        Dictionary containing averages over the training set.

    Returns
    -------
    torch.Tensor
        Scaled input tensor.
    """
    return h * (avg_d["log"] / np.log(D + 1))
